@@ -4,7 +4,7 @@
 
 ;; Author: Alex Harsanyi <AlexHarsanyi@gmail.com>
 ;; Created: 27 Sep 2006
-;; Version: 1.4
+;; Version: 1.5
 ;; Keywords: languages sql
 ;; Homepage: https://github.com/alex-hhh/emacs-sql-indent
 ;; Package-Requires: ((cl-lib "0.5"))
@@ -282,7 +282,7 @@ determine the statement start in SQLite scripts.")
    (regexp-opt '("use" "go" "declare") t)
    "\\)\\b")
   "Match an MS SQL Sever directive at the beginning of a line.")
-  
+
 (defun sqlind-beginning-of-directive ()
   "Return the position of an SQL directive, or nil.
 We will never move past one of these in our scan.  We also assume
@@ -309,7 +309,7 @@ But don't go before LIMIT."
     (catch 'done
       (while (> (point) (or limit (point-min)))
         (when (re-search-backward
-               ";\\|:=\\|\\_<\\(declare\\|begin\\|cursor\\|for\\|while\\|loop\\|if\\|then\\|else\\|elsif\\|elseif\\)\\_>\\|)"
+               ";\\|:=\\|\\_<\\(declare\\|begin\\|cursor\\|for\\|while\\|loop\\|if\\|then\\|else\\|elsif\\|elseif\\)\\_>\\|)\\|\\$\\$"
                limit 'noerror)
           (unless (sqlind-in-comment-or-string (point))
             (let ((candidate-pos (match-end 0)))
@@ -318,9 +318,14 @@ But don't go before LIMIT."
                      ;; of the keywords inside one of them and think this is a
                      ;; statement start.
                      (progn (forward-char 1) (forward-sexp -1)))
-                    ((looking-at "cursor\\|for\\|while")
-                     ;; statement begins at the start of the keyword
+                    ((looking-at "cursor\\|for")
+                     (unless (eq sql-product 'postgres)
+                       (throw 'done (point))))
+                    ((looking-at "while")
                      (throw 'done (point)))
+                    ((looking-at "declare")
+                     (when (eq sql-product 'postgres)
+                       (throw 'done (point))))
                     ((looking-at "else?if")
                      ;; statement begins at the start of the keyword
                      (throw 'done (point)))
@@ -340,6 +345,10 @@ But don't go before LIMIT."
                      (sqlind-backward-syntactic-ws)
                      (forward-sexp -1)
                      (throw 'done (point)))
+                    ((looking-at "\\$\\$")
+                     (when (eq sql-product 'postgres)
+                       (sqlind-forward-syntactic-ws)
+                       (throw 'done (point))))
                     ((sqlind-looking-at-begin-transaction)
                      ;; This is a "begin transaction" call, statement begins
                      ;; at "begin", see #66
@@ -488,7 +497,7 @@ We try to avoid false positives, like \"end if\" or the various
          (forward-word -1)
          ;; we don't want to match an "end if", and things like "drop index if
          ;; exists..." and "create index if not exist..."
-         (not (looking-at "end\\|table\\|view\\|index\\|trigger\\|procedure\\|function\\|package\\|body")))))
+         (not (looking-at "end\\|schema\\|table\\|view\\|index\\|trigger\\|procedure\\|function\\|package\\|body")))))
 
 (defun sqlind-maybe-if-statement ()
   "If (point) is on an IF statement, report its syntax."
@@ -634,11 +643,20 @@ expressions, also we don't match DECLARE directives here.
 
 See also `sqlind-beginning-of-block'"
   (when (looking-at "declare")
-    ;; a declare block is always toplevel, if it is not, its an error
-    (throw 'finished
-      (if (null sqlind-end-stmt-stack)
-	  'declare-statement
-        (list 'syntax-error "nested declare block" (point) (point))))))
+
+    ;; In Postgres, a DECLARE statement can be a block, or define a cursor
+    ;; (see pr67.sql, pr92.sql for examples).  It is somewhat tricky to
+    ;; determine which is which, so we use the heuristic that a declare
+    ;; statement immediately following a $$ is a block, otherwise it is not.
+    (when (or (not (eq sql-product 'postgres))
+              (save-excursion
+                (sqlind-backward-syntactic-ws)
+                (skip-syntax-backward "_w") ; note that the $$ is symbol constituent!
+                (looking-at "\\(\\$\\$\\)\\|begin\\|then\\|else")))
+      (throw 'finished
+        (if (null sqlind-end-stmt-stack)
+            'declare-statement
+          (list 'syntax-error "nested declare block" (point) (point)))))))
 
 (defun sqlind-maybe-skip-create-options ()
   "Move point past any MySQL option declarations.
@@ -721,9 +739,14 @@ present."
             (sqlind-forward-syntactic-ws))
            (t (throw 'finished nil))))))))
 
-(defun sqlind-maybe-create-statement ()
+(defun sqlind-maybe-create-statement (&optional all-statements)
   "If (point) is on a CREATE statement, report its syntax.
-See also `sqlind-beginning-of-block'"
+See also `sqlind-beginning-of-block'
+
+Normally, only block start create statements are considered (such
+as creation of procedures).  In particular, create
+table/view/index statements are ignored unless the ALL-STATEMENTS
+argument is t"
   (when (or (looking-at "create\\_>\\(?:[ \t\n\r\f]+\\)\\(or\\(?:[ \t\n\r\f]+\\)replace\\_>\\)?")
             (looking-at "alter\\_>"))
     (prog1 t                            ; make sure we return t
@@ -755,36 +778,40 @@ See also `sqlind-beginning-of-block'"
           (when (string-match "\\(.*?\\)(" name)
             (setq name (match-string 1 name)))
 
-	  (if (memq what '(procedure function package package-body))
-	      ;; check is name is in the form user.name, if so then suppress user part.
-	      (progn
-		(when (string-match "\\(?:.*\\.\\)?\\(.*\\)" name)
-		  (setq name (match-string 1 name)))
-		(if (null sqlind-end-stmt-stack)
-		(throw 'finished
-		  (list (if (memq what '(procedure function)) 'defun-start what)
-			name))
-		(cl-destructuring-bind (pos kind label) (pop sqlind-end-stmt-stack)
-		  (when (not (eq kind nil))
-		    (throw 'finished
-		      (list 'syntax-error
-			    "bad closing for create block" (point) pos)))
-		  (unless (sqlind-labels-match label name)
-		    (throw 'finished
-		      (list 'syntax-error
-			    "label mismatch in create block" (point) pos))))))
-	    ;; we are creating a non-code block thing: table, view,
-	    ;; index, etc.  These things only exist at toplevel.
-	    (unless (null sqlind-end-stmt-stack)
-	      (throw 'finished
-		(list 'syntax-error "nested create statement" (point) (point))))
-	    (throw 'finished (list 'create-statement what name))))))))
+	  (cond
+            ((memq what '(procedure function package package-body))
+             ;; check is name is in the form user.name, if so then suppress user part.
+	     (when (string-match "\\(?:.*\\.\\)?\\(.*\\)" name)
+	       (setq name (match-string 1 name)))
+	     (if (null sqlind-end-stmt-stack)
+		 (throw 'finished
+		   (list (if (memq what '(procedure function)) 'defun-start what)
+			 name))
+	       (cl-destructuring-bind (pos kind label) (pop sqlind-end-stmt-stack)
+		 (when (not (eq kind nil))
+		   (throw 'finished
+		     (list 'syntax-error
+			   "bad closing for create block" (point) pos)))
+		 (unless (sqlind-labels-match label name)
+		   (throw 'finished
+		     (list 'syntax-error
+			   "label mismatch in create block" (point) pos))))))
+            ((memq what '(table view index))
+             ;; Table, view and index creations do not begin blocks and they
+             ;; are ignored unless the ALL-STATEMENTS parameter is t
+             (when all-statements
+               (throw 'finished (list 'create-statement what name))))
+            (t
+	     (unless (null sqlind-end-stmt-stack)
+	       (throw 'finished
+		 (list 'syntax-error "nested create statement" (point) (point))))
+	     (throw 'finished (list 'create-statement what name)))))))))
 
 (defun sqlind-maybe-defun-statement ()
   "If (point) is on a procedure definition statement, report its syntax.
 See also `sqlind-beginning-of-block'"
   (catch 'exit
-    (when (looking-at "\\(procedure\\|function\\)\\(?:[ \t\n\r\f]+\\)\\([a-z0-9_]+\\)")
+    (when (looking-at "\\(procedure\\|function\\)\\(?:[ \t\n\r\f]+\\)\\(?:[a-z0-9_]+\\.\\)?\\([a-z0-9_]+\\)")
       (prog1 t                          ; make sure we return t
 	(let ((proc-name (sqlind-match-string 2)))
 	  ;; need to find out if this is a procedure/function
@@ -803,13 +830,21 @@ See also `sqlind-beginning-of-block'"
 	      ;; not a procedure after all.
 	      (throw 'exit nil)))
 
-          ;; Find out if it is a drop procedure or function statement
           (save-excursion
             (sqlind-backward-syntactic-ws)
+            ;; Find out if it is a drop procedure or function statement
             (forward-word -1)
             (when (looking-at "drop")
               ;; not a procedure after all
-              (throw 'exit nil)))
+              (throw 'exit nil))
+            ;; Find out if it is a "comment on" statement (postgres only)
+            (when (and (eq sql-product 'postgres)
+                       (looking-at "on"))
+              (sqlind-backward-syntactic-ws)
+              (forward-word -1)
+              (when (looking-at "comment")
+                ;; not a procedure after all.
+                (throw 'exit nil))))
 
 	  ;; so it is a definition
 
@@ -868,6 +903,12 @@ See also `sqlind-beginning-of-block'"
                ;; which ends another statement)
                (push (list saved-pos '$$ "") sqlind-end-stmt-stack)
                (goto-char saved-pos))
+              ((progn (forward-word -1)
+                      (looking-at "end"))
+               ;; Assume the $$ is ending a statement (previous line contains
+               ;; an "end" keyword)
+               (push (list saved-pos '$$ "") sqlind-end-stmt-stack)
+               (goto-char saved-pos))
               ((null sqlind-end-stmt-stack)
                (sqlind-beginning-of-statement)
                (let ((syntax (catch 'finished
@@ -916,7 +957,7 @@ reverse order (a stack) and is used to skip over nested blocks."
   ;;
   ;; Some of these `sqlind-maybe-*` functions are specific to the
   ;; `sql-product` and are only invoked for the speficied SQL dialect.
-  
+
   (catch 'finished
     (let ((sqlind-end-stmt-stack end-statement-stack))
       (while (re-search-backward sqlind-start-block-regexp sqlind-search-limit 'noerror)
@@ -930,7 +971,8 @@ reverse order (a stack) and is used to skip over nested blocks."
             (sqlind-maybe-else-statement)
             (sqlind-maybe-loop-statement)
             (sqlind-maybe-begin-statement)
-            (when (eq sql-product 'oracle) ; declare statements only start blocks in PL/SQL
+            (when (memq sql-product '(oracle postgres))
+              ;; declare statements only start blocks in PL/SQL and PostgresSQL
               (sqlind-maybe-declare-statement))
             (when (eq sql-product 'postgres)
               (sqlind-maybe-$$-statement))
@@ -1107,7 +1149,7 @@ statement is found."
                            (sqlind-backward-syntactic-ws)
 			   (looking-at ",")))
 		 (throw 'finished (cons 'select-table match-pos)))
-               
+
 	       ;; otherwise, we continue the table definition from the
 	       ;; previous line.
 	       (throw 'finished (cons 'select-table-continuation match-pos)))
@@ -1397,7 +1439,7 @@ not a statement-continuation POS is the same as the
   (let ((syntax (sqlind-syntax context))
         (anchor (sqlind-anchor-point context))
         (syntax-symbol (sqlind-syntax-symbol context)))
-    
+
     (goto-char pos)
 
     (cond
@@ -1449,14 +1491,28 @@ not a statement-continuation POS is the same as the
          ;; CURSOR name type IS
          (when (looking-at "cursor\\b")
            (let ((origin (point)))
-             (forward-sexp 3)
+             (forward-sexp 1)           ; skip "cursor"
              (sqlind-forward-syntactic-ws)
-             (when (looking-at "is\\b")
-               (goto-char (match-end 0))
-               (sqlind-forward-syntactic-ws))
+             (forward-sexp 1)
+             (sqlind-forward-syntactic-ws)
+             (if (looking-at "is\\b")
+                 (progn
+                   (goto-char (match-end 0))
+                   (sqlind-forward-syntactic-ws))
+               (forward-sexp 1)
+               (sqlind-forward-syntactic-ws)
+               (when (looking-at "is\\b")
+                 (goto-char (match-end 0))
+                 (sqlind-forward-syntactic-ws)))
              (unless (<= (point) pos)
                (goto-char origin))))
 
+         ;; Skip a PostgreSQL cursor declaration
+         (when (and (eq sql-product 'postgres)
+                    (looking-at "\\(\\(declare\\)\\|\\(cursor\\)\\|\\(for\\)\\)\\b"))
+           (when (re-search-forward "\\b\\(select\\|update\\|delete\\|insert\\)\\b" pos 'noerror)
+             (goto-char (match-beginning 0))))
+
          ;; skip a forall statement if it is before our point
          (when (looking-at "forall\\b")
            (when (re-search-forward "\\b\\(select\\|update\\|delete\\|insert\\)\\b" pos 'noerror)
@@ -1578,6 +1634,15 @@ not a statement-continuation POS is the same as the
      (push (sqlind-refine-end-syntax
             nil "" (point) context)
            context))
+
+    ;; See #92 and pr92b.sql, no such thing as a nested declare statement, use
+    ;; the context of the previous declare-statement!
+    ((and (eq sql-product 'postgres)
+          (eq syntax-symbol 'declare-statement)
+          (looking-at "declare\\_>"))
+     (goto-char anchor)
+     (setq context (sqlind-syntax-of-line)))
+
     )
   context))
 
@@ -1604,6 +1669,13 @@ procedure block."
         (goto-char context-start)
         (when (or (>= context-start pos)
                   (and (looking-at sqlind-start-block-regexp)
+                       ;; create table/view/index statements are not block
+                       ;; contexts
+                       (or (not (looking-at "\\(create\\)\\|\\(alter\\)"))
+                           (catch 'finished (sqlind-maybe-create-statement) nil))
+                       ;; A declare statement may or may not be a block context
+                       (or (not (looking-at "declare"))
+                           (catch 'finished (sqlind-maybe-declare-statement) nil))
                        (not (sqlind-looking-at-begin-transaction))))
           (goto-char pos)
           ;; if we are at the start of a statement, or the nearest statement
@@ -1626,6 +1698,14 @@ procedure block."
               (setq context-start (point))
               (setq context (list (cons block-info context-start))))))
 
+        (goto-char context-start)
+        (when (and (eq 'statement-continuation (sqlind-syntax-symbol context))
+                   (looking-at "\\(create\\)\\|\\(alter\\)"))
+          (let ((create-info (catch 'finished (sqlind-maybe-create-statement t) nil)))
+            (when create-info
+              (pop context)         ; remove the statement continuation syntax
+              (push (cons create-info context-start) context))))
+
         (let ((parse-info (syntax-ppss pos)))
           (cond ((nth 4 parse-info)   ; inside a comment
                  (push (cons 'comment-continuation (nth 8 parse-info)) context))
@@ -1866,11 +1946,14 @@ clause (select, from, where, etc) in which the current point is.
     (in-select-clause               sqlind-lineup-to-clause-end
                                     sqlind-right-justify-logical-operator)
     (insert-clause                  sqlind-right-justify-clause)
-    (in-insert-clause               sqlind-lineup-to-clause-end)
+    (in-insert-clause               sqlind-lineup-to-clause-end
+                                    sqlind-right-justify-logical-operator)
     (delete-clause                  sqlind-right-justify-clause)
-    (in-delete-clause               sqlind-lineup-to-clause-end)
+    (in-delete-clause               sqlind-lineup-to-clause-end
+                                    sqlind-right-justify-logical-operator)
     (update-clause                  sqlind-right-justify-clause)
-    (in-update-clause               sqlind-lineup-to-clause-end))
+    (in-update-clause               sqlind-lineup-to-clause-end
+                                    sqlind-right-justify-logical-operator))
   "Define the indentation amount for each syntactic symbol.
 
 The value of this variable is an ALIST with the format:
@@ -2130,7 +2213,10 @@ with AND, OR or NOT to be aligned so they sit under the WHERE clause."
     (cl-destructuring-bind ((_sym clause) . anchor) (car syntax)
       (if (and (equal clause "where")
                (looking-at "and\\|or\\|not"))
-          (- base-indentation (1+ (- (match-end 0) (match-beginning 0))))
+          (progn
+            (goto-char anchor)
+            (+ (current-column)
+               (- (length clause) (- (match-end 0) (match-beginning 0)))))
         base-indentation))))
 
 (defun sqlind-left-justify-logical-operator (syntax base-indentation)
